local _, BigBrother = ...

local module = BigBrother:NewModule("Alerts", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BigBrother")

local _G = _G
local bit_band = _G.bit.band
local bit_bor = _G.bit.bor

local UNKNOWN = ("(%s)"):format(_G.UNKNOWN)

--localize names for reverse lookup
local crowdControlSpells = {}
for _,v in ipairs(BigBrother.SpellData.crowdControl) do
	crowdControlSpells[(GetSpellInfo(v))] = true
end

local misdirectSpells = {}
for _,v in ipairs(BigBrother.SpellData.misdirect) do
	misdirectSpells[v] = true
end

local tauntSpells = {}
local tauntAoESpells = {}
for _,v in ipairs(BigBrother.SpellData.taunt) do
	tauntSpells[(GetSpellInfo(v))] = true
end
for _,v in ipairs(BigBrother.SpellData.tauntAoE) do
	tauntAoESpells[(GetSpellInfo(v))] = true
end

local interruptSpells = {}
for _,v in ipairs(BigBrother.SpellData.interrupts) do
	interruptSpells[(GetSpellInfo(v))] = true
end


function module:OnInitialize()
	self.db = BigBrother.db:RegisterNamespace("Alerts", {
		profile = {
			crowdControl = true,
			misdirect = true,
			taunt = false,
			interrupt = false,
			combatRes = true,
			spellLink = true,
			groupOnly = true,
			icons = true,
			output = "self",
			outputExtra = "",
		}
	})
end

function module:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("CHANNEL_UI_UPDATE")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "CHANNEL_UI_UPDATE")
end


local function GetIconString(flags)
	if module.db.profile.icons then
		local raidTarget = bit_band(flags, _G.COMBATLOG_OBJECT_RAIDTARGET_MASK)
		if raidTarget ~= 0 then
			for i=1,8 do
				local mask = _G.COMBATLOG_OBJECT_RAIDTARGET1 * (2 ^ (i - 1)) --_G["COMBATLOG_OBJECT_RAIDTARGET"..i]
				if raidTarget == mask then
					return _G.TEXT_MODE_A_STRING_DEST_ICON:format(mask, _G["COMBATLOG_ICON_RAIDTARGET"..i])
				end
			end
		end
	end
	return ""
end


function module:Spam(msg)
	if not msg or msg == "" then return end
	local chatMsg = string.gsub(msg, "|Hicon:%d+:dest|h|TInterface.TargetingFrame.UI%-RaidTargetingIcon_(%d).blp:0|t|h", "{rt%1}")
	chatMsg = string.gsub(chatMsg,"|c%x%x%x%x%x%x%x%x([%w%s]-)|r", "%1") --colored text in SendChatMessage can disconnect the user, apparently (spell links are special)
	local output = self.db.profile.output
	
	if output == "self" then
		BigBrother:Print(msg)
	elseif output == "group" then
		if GetNumRaidMembers() > 0 then
			SendChatMessage(chatMsg, "RAID")
		elseif GetNumPartyMembers() > 0 then
			SendChatMessage(chatMsg, "PARTY")
		end
	elseif output == "raid" then
		if GetNumRaidMembers() > 0 then
			SendChatMessage(chatMsg, "RAID")
		end
	elseif output == "party" then
		if GetNumPartyMembers() > 0 then
			SendChatMessage(chatMsg, "PARTY")
		end
	elseif output == "guild" then
		if IsInGuild() and (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) then
			SendChatMessage(chatMsg, "GUILD")
		end
	elseif output == "officer" then
		if IsInGuild() and (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) then
			SendChatMessage(chatMsg, "OFFICER")
		end
	elseif output == "channel" then
		local index = GetChannelName(self.db.profile.outputExtra)
		if index > 0 then
			SendChatMessage(chatMsg, "CHANNEL", nil, index)
		end
	end
end

do
	local spamCount = 0
	local spamLastTime = 0
	
	local FILTER_GROUP = bit_bor(COMBATLOG_OBJECT_TYPE_PLAYER, COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_AFFILIATION_RAID)
	local function checkSrc(flags)
		if module.db.profile.groupOnly then
			return flags and (bit_band(flags, FILTER_GROUP) ~= 0)
		end
		return true
	end
	
	function module:COMBAT_LOG_EVENT_UNFILTERED(_, _, event, _, srcName, srcFlags, _, dstName, dstFlags, spellId, spellName, _, ...)
		
		--CC Breaking
		if self.db.profile.crowdControl
			and (event == "SPELL_AURA_BROKEN" or event == "SPELL_AURA_BROKEN_SPELL"
				or (event == "SPELL_AURA_REMOVED" and spellId == 51514)) --Hex
			and crowdControlSpells[spellName]
			and checkSrc(srcFlags)
			and (bit_band(dstFlags, _G.COMBATLOG_OBJECT_REACTION_HOSTILE) ~= 0) then
			
			local srcOutput = ("%s|cff40ff40%s|r"):format(GetIconString(srcFlags), srcName or UNKNOWN)
			local dstOutput = ("%s|cffff4040%s|r"):format(GetIconString(dstFlags), dstName or UNKNOWN)
			local spellOutput = self.db.profile.spellLink and GetSpellLink(spellId) or spellName
			
			local output
			if event == "SPELL_AURA_BROKEN" or event == "SPELL_AURA_REMOVED" then
				output = (L["%s on %s removed by %s"]):format(spellOutput, dstOutput, srcOutput)
			elseif event == "SPELL_AURA_BROKEN_SPELL" then
				local extraSpellId, extraSpellName = ...
				local extraOutput = self.db.profile.spellLink and GetSpellLink(extraSpellId) or extraSpellName
				output = (L["%s on %s removed by %s's %s"]):format(spellOutput, dstOutput, srcOutput, extraOutput)
			end
			
			--throttling...hmmm
			local now = GetTime()
			if now - spamLastTime > 3 then
				spamCount = 0
			end
			spamCount = spamCount + 1
			if spamCount > 3 then
				output = nil --("%s is being bad"):format(srcOut) --
			end
			spamLastTime = now
			
			self:Spam(output)
			return
		end
		
		--Misdirects
		if self.db.profile.misdirect and event == "SPELL_CAST_SUCCESS" and misdirectSpells[spellId] and checkSrc(srcFlags) then
			local srcOutput = ("%s|cff40ff40%s|r"):format(GetIconString(srcFlags), srcName or UNKNOWN)
			local dstOutput = ("%s|cffff4040%s|r"):format(GetIconString(dstFlags), dstName or UNKNOWN)
			local spellOutput = self.db.profile.spellLink and GetSpellLink(spellId) or spellName
			
			self:Spam(L["%s cast %s on %s"]:format(srcOutput, spellOutput, dstOutput))
			return
		end
		
		--Interrupts
		if self.db.profile.interrupt and (event == "SPELL_INTERRUPT" or (event == "SPELL_MISSED" and interruptSpells[spellName])) and checkSrc(srcFlags) then
			local srcOutput = ("%s|cff40ff40%s|r"):format(GetIconString(srcFlags), srcName or UNKNOWN)
			local dstOutput = ("%s|cffff4040%s|r"):format(GetIconString(dstFlags), dstName or UNKNOWN)
			local spellOutput = self.db.profile.spellLink and GetSpellLink(spellId) or spellName
			
			local output
			if event == "SPELL_INTERRUPT" then --doesn't trigger on silence/stun :\ oh well
				local extraSpellId, extraSpellName = ... --the interrupted spell
				local extraSpellOutput = self.db.profile.spellLink and GetSpellLink(extraSpellId) or extraSpellName
				output = L["%s interrupted %s's %s with %s"]:format(srcOutput, dstOutput, extraSpellOutput, spellOutput)
			elseif event == "SPELL_MISSED" then
				local missType = ...
				output = L["%s cast %s on %s (%s)"]:format(srcOutput, spellOutput, dstOutput, _G["ACTION_SPELL_MISSED_"..missType])
			end
			
			self:Spam(output)
			return
		end
		
		--Taunts
		if self.db.profile.taunt and (
			((event == "SPELL_AURA_APPLIED" and tauntSpells[spellName]) or
					(event == "SPELL_MISSED" and tauntSpells[spellName]) or
					(event == "SPELL_CAST_SUCCESS" and tauntAoESpells[spellName]))
				and checkSrc(srcFlags)
			) then
			
			local srcOutput = ("%s|cff40ff40%s|r"):format(GetIconString(srcFlags), srcName or UNKNOWN)
			local dstOutput = ("%s|cffff4040%s|r"):format(GetIconString(dstFlags), dstName or UNKNOWN)
			local spellOutput = self.db.profile.spellLink and GetSpellLink(spellId) or spellName
			
			local output
			if event == "SPELL_AURA_APPLIED" then
				output = L["%s cast %s on %s"]:format(srcOutput, spellOutput, dstOutput)
			elseif event == "SPELL_MISSED" then
				local missType = ...
				if spellName == (GetSpellInfo(49576)) and missType == "IMMUNE" then return end --ignore dg yoink immunes
				output = L["%s cast %s on %s (%s)"]:format(srcOutput, spellOutput, dstOutput, _G["ACTION_SPELL_MISSED_"..missType])
			elseif event == "SPELL_CAST_SUCCESS" then
				output = L["%s cast %s"]:format(srcOutput, spellOutput)
			end
			
			self:Spam(output)
			return
		end
		
		--Battle Res
		if self.db.profile.combatRes and event == "SPELL_RESURRECT" and spellId == 20484 and checkSrc(srcFlags) then
			local srcOutput = ("%s|cff40ff40%s|r"):format(GetIconString(srcFlags), srcName or UNKNOWN)
			local dstOutput = ("%s|cffff4040%s|r"):format(GetIconString(dstFlags), dstName or UNKNOWN)
			local spellOutput = self.db.profile.spellLink and GetSpellLink(spellId) or spellName
			
			self:Spam(L["%s cast %s on %s"]:format(srcOutput, spellOutput, dstOutput))
			return
		end
		
	end
end

local GetExtras
do
	local channels = {}
	local function updateChannels()
		wipe(channels)
		for i = 1, GetNumDisplayChannels() do
			local name, _, _, _, _, _, category = GetChannelDisplayInfo(i)
			if category == "CHANNEL_CATEGORY_CUSTOM" then
				channels[name] = name
			end
		end
	end
	
	--just channels for now. should probably scavenge LibSink some more for sct support
	function GetExtras()
		updateChannels()
		return channels
	end
	BigBrother.GetChatChannels = GetExtras
end

function module:CHANNEL_UI_UPDATE()
	LibStub("AceConfigRegistry-3.0"):NotifyChange("BigBrother")
end

function module:AddOptions(options)
	options.args.alerts = {
		name = L["Alerts"],
		--desc = L["Alerts"],
		type = "group",
		get = function(info) return self.db.profile[info[#info]] end,
		set = function(info, value) self.db.profile[info[#info]] = value end,
		order = 20,
		args = {
			report = {
				name = L["Report"],
				desc = L["Set what is reported to chat"],
				type = "group",
				inline = true,
				order = 10,
				args = {
					crowdControl = {
						name  = L["Crowd Control"],
						desc = L["Report when a player breaks a crowd control effect"],
						type = "toggle",
						order = 5,
					},
					misdirect = {
						name  = L["Misdirect"],
						desc = L["Report who gains Misdirection"],
						type = "toggle",
						order = 10,
					},
					taunt = {
						name  = L["Taunt"],
						desc = L["Report taunts"],
						type = "toggle",
						order = 15,
					},
					interrupt = {
						name  = L["Interrupts"],
						desc = L["Report interrupts"],
						type = "toggle",
						order = 16,
					},
					combatRes = {
						name = L["Combat Resurrection"],
						desc = L["Report battle resurrection casts"],
						type = "toggle",
						order = 17,
					},
				},
			}, --report
			
			output = {
				name = L["Output"],
				desc = L["Set where the alert output is sent"],
				type = "group",
				inline = true,
				get = function(info)
					local key = info[#info]
					return self.db.profile.output == key
				end,
				set = function(info, value)
					if value then
						local key = info[#info]
						self.db.profile.output = key
					end
				end,
				order = 20,
				args = {
					self = {
						name = L["Self"],
						desc = L["Reports alerts to yourself"],
						type = "toggle",
						order = 5,
					},
					group = {
						name = L["Group"],
						desc = L["Reports alerts to party or raid chat"],
						type = "toggle",
						order = 10,
					},
					party = {
						name = L["Party"],
						desc = L["Reports alerts to party chat"],
						type = "toggle",
						order = 11,
					},
					raid = {
						name = L["Raid"],
						desc = L["Reports alerts to raid chat"],
						type = "toggle",
						order = 12,
					},
					guild = {
						name = L["Guild"],
						desc = L["Reports alerts to guild chat"],
						type = "toggle",
						order = 15,
					},
					officer = {
						name = L["Officer"],
						desc = L["Reports alerts to officer chat"],
						type = "toggle",
						order = 20,
					},
					channel = {
						name = L["Channel"],
						desc = L["Reports alerts to a custom channel"],
						type = "toggle",
						order = 25,
					},
					spacer = {
						name = " ", type = "description", order = 100,
						hidden = function() return self.db.profile.output ~= "channel" end,
					},
					extra = {
						name = "",
						type = "select",
						values = GetExtras,
						get = function() return self.db.profile.outputExtra end,
						set = function(info, value) self.db.profile.outputExtra = value end,
						hidden = function() return self.db.profile.output ~= "channel" end,
						order = 101,
					},
				},
			},
			
			spellLink = {
				name = L["Use Spell Links"],
				desc = L["Display spell names as clickable spell links"],
				type = "toggle",
				order = 40,
			},
			groupOnly = {
				name = L["Group Only"],
				desc = L["Only report events from players in your party or raid group"],
				type = "toggle",
				order = 41,
			},
			icons = {
				name = L["Raid Target Icons"],
				desc = L["Show associated raid target icons as part of the alert"],
				type = "toggle",
				order = 42,
			},
			
		},
	}
end
